--script.on_event(defines.events.on_player_removed_equipment,function(e) print(e.equipment)  end )
--do return  end

--some constants that may need updating after some release:
energy_per_shot = 1000 --constant related to amount of energy used in a single shot
check_interval = 30 --interval between checks of players
type_to_defines = {--correspondence between entity type and main inventory from which stuff should be pulled
	['player'] = defines.inventory.player_main,
	['car'] = defines.inventory.car_trunk,
	['cargo-wagon'] = defines.inventory.cargo_wagon,
}

--Functions are grouped by their call depth, first are those that are directly linked to game, then their auxiliaries then auxiliaries' auxiliaries, hope that helps whomever is trying to read this
--[[
	Level 1
]]
function init()
	--creation of information tables about generated entities
	------------
	--These tables will be used:
	global.equip_data = {--[[
		[equipment_name] = {
			ammo = ammo_name,
			magazine_size = int,
			magazine_energy = number,
			ammo_category = {
				[1] = ammo_name,
			}
		}
	]]}
	global.ammo_equip = {--[[
		[ammo_name] = equipment_name
	]]}
		--global table to store all the characters that should be updated
	global.bodies = global.bodies or {
	--[[
		[unit_number] = {
			inventory = inventory_reference,
			self = character_reference,
			grid = reference_of_armor_grid,
			[1] = {
				handle,--these fields were once [1],[2],[3] 
				energy,--but I doubt the speed gains of that would offset the time wasted deciphering the code
				name,--with those, especially since I assign them to local variables when in use
			},
		}
	]]
	} 
	global.update = global.update or {--a table to simplify update spreading over ticks
		--[[
			[index] = o
		]]
	}
	-----------Here we'll initialize them
	local ammo2category = {--[[
		[ammo_name] = ammo_category
	]]}
	local category2ammo = {--[[
		[ammo_category] = {
			[1] = ammo_name,
		}
	]]}
	for k,v in pairs(game.item_prototypes) do
		if string.find(k,"automated%-equipment%-t#") then
		--equipment is named as 'automated-equipment-#' ..ammo.name.. '#'
			local category,ammo
			_,_,category,ammo = string.find(k,"t#(.+)#n#(.+)#")
			--first pass on filling the tables
			global.equip_data[k] = {
				ammo = ammo,
				magazine_size = game.item_prototypes[ammo].magazine_size,
				magazine_energy = game.item_prototypes[ammo].magazine_size * energy_per_shot,
				}
			global.ammo_equip[ammo] = k
			--auxiliary table
			ammo2category[ammo] = category
		end
	end
	--make tables containing bullets of the same category
	for k,v in pairs(ammo2category) do
		category2ammo[v] = category2ammo[v] or {}
		table.insert(category2ammo[v],k)
	end
	--find every equipment that uses bullets of said category and link the category table there
	for cat,cat_tab in pairs(category2ammo) do
		for _, bullet_name in pairs(cat_tab) do
			global.equip_data[global.ammo_equip[bullet_name]].category = cat_tab
		end
	end
end

function placed_entity(event)
	--some entity is placed and may have the grid in them
	local e = event.created_entity
	local g
	local i
	if e and e.valid then
		g = e.grid and e.grid.valid and e.grid
		i = e.get_inventory(type_to_defines[e.type])
	end
	
	if g and i then check_new_grid(e,i,g) end
end

function placed_armor(event)
	--player's armor is worn
	
	local p = game.players[event.player_index]
	if not (p.valid and p.connected )  then return end
	
	local c = p.character
	if not (c and c.valid) then return end
	
	local i
	local armor = c.get_inventory(defines.inventory.player_armor)[1]
	local g= armor and armor.valid and armor.valid_for_read and armor.grid
	if not (g and g.valid) then return end
	
	check_new_grid(c,i,g)
end

function grid_put(event)
	--grid is getting new equipment, need to add it to tracker
	
	local eq = event.equipment
	if not (eq and eq.valid and global.equip_data[eq.name]) then return end
	
	local p = game.players[event.player_index]
	if not (p.valid and p.connected)  then return end
	
	local grid = event.grid
	if not (grid and grid.valid) then return  end
	
	local body,inv = find_relevant(p,grid)
	if not body then return end
	insert_equipment(body,inv,grid,eq)
end

function update( event )--this is on_tick update, it better be quick
	local t = event.tick
	local u = global.update
	
	for i = #u-t%30, 1,-30 do--this spreads update so that one per 30 bodies is updated in tick
		local o = u[i]
		if not (o.self.valid and o.grid.valid and o.inventory.valid) then --validity checks
			global_iremove(i) 
		else--update equipments carried by this body
			local equipment_data = global.equip_data
			local per_shot = energy_per_shot
		
			for j = #o,1,-1 do
				local object = o[j]--equipment "object" (data I use when update equipment ingame object)
				local e = object.handle
				if not (e and e.valid) then remove_equipment(o,j) else--check validity, then
					local ammo_data = equipment_data[object.name]
					local menergy = ammo_data.magazine_energy
					
					if e.max_energy - e.energy > menergy then--see if reload is required
						local inv = o.inventory
						local ammo = ammo_data.ammo
						if inv.get_item_count(ammo) > 10 then--is there ammo
							inv.remove{name = ammo,count = 1}
							e.energy = e.energy+ menergy
						elseif  e.energy < per_shot then--or is ammo critically low
							
							--is there other ammo of same category
							local category = ammo_data.category
							
							--is there other ammo of same category in this inventory
							local ammo
							for i = #category,1,-1 do
								local a = category[i]
								if inv.get_item_count(a) > 10 then 
									ammo = a
									break
								end 
							end
							
							if ammo then--if there indeed is
								--swap to equivalent equipment
								local grid = o.grid
								local pos = e.position
								grid.take{equipment = e}
								e = grid.put{name = global.ammo_equip[ammo],position = pos}
								--load that one
								inv.remove{name = ammo,count = 1}
								e.energy = e.energy+ equipment_data[e.name].magazine_energy
								--put new references in `eq` object
								object.handle = e
								object.name = e.name
							end
						end
					end
					object.energy = e.energy
				end
			end
		end
	end
end

function grid_unput(event)
	--grid is losing old equipment, need to return ammo
	
	if not (global.equip_data[event.equipment]) then return end
	
	local p = game.players[event.player_index]
	if not (p.valid and p.connected)  then return end
	
	local g = event.grid
	
	local body,inv = find_relevant(p,g)
	if not body then return end
	
	local o = global.bodies[body.unit_number]
	if not o then return end
	
	for i = #o,1,-1 do
		if not o[i].handle.valid then
			remove_equipment(o,i)
		end
	end
end

--[[
	Level 2
]]
function check_new_grid(body,inv,grid)
	--store new body to be updated
	for _,eq in pairs(grid.equipment) do
		if eq.valid and global.equip_data[eq.name] then
			insert_equipment(body,inv,grid,eq)
		end
	end
end

function remove_equipment(o, index)--remove equipment from update list
	local edata = global.equip_data
	local energy_per_shot = energy_per_shot
	
	local e= o[index]
	local ammo_data = edata[e.name]
	local ammo_name = ammo_data.ammo
	local mag_energy = ammo_data.magazine_energy
	local count = (e.energy/mag_energy)
	if count ~= 0 then
		local remainder = (count%1)
		count = count - remainder
		o.inventory.insert{name = ammo_name,count = count}
		--if remainder~ = 0 then --when loading, I cant tell half filled magazine from full, so if I'd give non-full magazines back that'd be exploit
			--o.inventory.insert{name = ammo_name,count = 1,ammo = remainder}
		--end
	end
	table.remove(o,index)
	if #o == 0 then 
		global_remove(o)
	end
end

function find_relevant(player,grid)--find entity and inventory the player is tampering with
	--find inventory to pull from
	local inventory
	local body
	--first guess - player's armor and body
	local ch = player.character
	local inv1 = ch and ch.valid and ch.get_inventory(type_to_defines[ch.type])
	if inv1 then
		local armor = ch.get_inventory(defines.inventory.player_armor)[1]
		local g= armor and armor.valid and armor.valid_for_read and armor.grid
		if g and g == grid then 
			inventory = inv1
			body = ch
		end
	end
	--second guess - some entity the player is tampering
	local op = player.opened
	if op and op.valid and op.grid and op.grid == grid then 
		inventory = op.get_inventory(type_to_defines[op.type])
		body = op
	end
	
	return body,inventory
end

--[[
	Level 3
]]
function insert_equipment(body,inv,grid,eqv)
	--store in the global the information for updating the state of equipments
	
	local bodies = global.bodies
	
	local o = global_get(body)  or global_add(body,inv,grid)

	for i = 1,#o do if o[i].handle == eqv then return end end --check for duplicates
	
	table.insert(o,{handle = eqv, name = eqv.name, energy = eqv.energy})
end

function global_get(body)--these global_* functions are somewhat redundant but they might be needed for various interfaces
	return global.bodies[body.unit_number]
end

function global_add(body,inv,grid)
	local o = global.bodies[body.unit_number]
	if not o then
		o = {
		uid = body.unit_number,
		inventory = inv,
		self = body,
		grid = grid,
		}
		global.bodies[o.uid] = o
		table.insert(global.update,o)
	end
	return o
end

function global_remove(o)
	global.bodies[o.uid] = nil
	local u = global.update
	for i = #u,1,-1 do
		if u[i] ==o then table.remove(u,i) end
	end
end

function global_iremove(i)
	global.bodies[ global.update[i].uid ] = nil
	table.remove(global.update,i)
end

function migrate_from_0_1_5( )---migration
	if global.characters then
		local gc = global.characters
		local bodies = global.bodies
		for i = 1,#gc do
			local c = gc[i]
			if c 
				and c.self and c.self.valid 
				and c.inventory and c.inventory.valid 
				and c.grid and c.grid.valid
			then
				check_new_grid(c.self,c.inventory,c.grid)
			end
		end
		global.characters = nil
		global.next = nil
	end
end
------------------------------------------------------
--Debug interface
remote.add_interface("Advanced_Personal_Debug",{
	--debug interface
	ammo_equip = function()
		print(serpent.block(global.ammo_equip))
	end,
	equip_data = function()
		print(serpent.block(global.equip_data))
	end,
	update = function()
		print(serpent.block(global.update))
	end,
})
-----------------------------------------------------
--Bug interface for AAI
remote.add_interface("Advanced_Personal_Defence_replacement", { 
	on_entity_replaced = function(event)  
		placed_entity{created_entity=event.new_entity}
	end
})
----------------------------------------------------
--handler registration
do
local de = defines.events
script.on_event(de.on_player_placed_equipment,grid_put)
script.on_event(de.on_player_removed_equipment,grid_unput)
script.on_event({de.on_built_entity, de.on_robot_built_entity},placed_entity)
script.on_event(de.on_player_armor_inventory_changed,placed_armor)
script.on_event(de.on_tick,update)
script.on_init(function() init() end)
script.on_configuration_changed(function(change) 
	init() 
	migrate_from_0_1_5()
	end)
end
